from ape.api import PluginConfig


class RateLimitConfig(PluginConfig):
    """
    Configuration for rate limiting.

    Args:
        min_retry_delay (int): The amount of milliseconds to wait before
          retrying the request. Defaults to ``1_000`` (one second).
        retry_backoff_factor (int): The multiplier applied to the retry delay
          after each failed attempt. Defaults to ``2``.
        max_retry_delay (int): The maximum length of the retry delay.
          Defaults to ``30_000`` (30 seconds).
        max_retries (int): The maximum number of retries.
          Defaults to ``3``.
        retry_jitter (int): A random number of milliseconds up to this limit
          is added to each retry delay. Defaults to ``250`` milliseconds.
    """

    min_retry_delay: int = 1_000
    retry_backoff_factor: int = 2
    max_retry_delay: int = 30_000
    max_retries: int = 3
    retry_jitter: int = 250


class AlchemyConfig(PluginConfig):
    """
    Configuration for Alchemy.

    Args:
        rate_limit (RateLimitConfig): The rate limiting configuration.
        trace_timeout (int): The maximum amount of milliseconds to wait for a
          trace. Defaults to ``10_000`` (10 seconds).
    """

    rate_limit: RateLimitConfig = RateLimitConfig()
    trace_timeout: str = "10s"



import os
from typing import TYPE_CHECKING

import ape
import pytest
from requests import HTTPError, Response
from web3 import Web3

if TYPE_CHECKING:
    from ape_alchemy.provider import Alchemy

FEATURE_NOT_AVAILABLE_BECAUSE_OF_TIER_RESPONSE = (
    "trace_transaction is not available on the Free tier - "
    "upgrade to Growth or Enterprise for access. See available methods at "
    "https://docs.alchemy.com/alchemy/documentation/apis"
)
FEATURE_NOT_AVAILABLE_BECAUSE_OF_NETWORK_RESPONSE = (
    "trace_transaction is not available on the ETH_SEPOLIA. "
    "For more information see our docs: "
    "https://docs.alchemy.com/alchemy/documentation/apis/ethereum"
)


@pytest.fixture
def accounts():
    return ape.accounts


@pytest.fixture
def networks():
    return ape.networks


@pytest.fixture
def missing_token(alchemy_provider, mocker):
    env = os.environ.copy()
    mock = mocker.patch("os.environ.get")

    def side_effect(key, *args, **kwargs):
        return None if "WEB3" in key else env.get(key, *args, **kwargs)

    alchemy_provider.network_uris = {}
    mock.side_effect = side_effect
    return mock


@pytest.fixture
def token(mocker):
    env = os.environ.copy()
    mock = mocker.patch("os.environ.get")

    def side_effect(key, *args, **kwargs):
        return "TEST_TOKEN" if "WEB3" in key else env.get(key, *args, **kwargs)

    mock.side_effect = side_effect
    return mock


@pytest.fixture
def mock_web3(mocker):
    mock = mocker.MagicMock(spec=Web3)
    mock.eth = mocker.MagicMock()
    mock.manager = mocker.MagicMock()
    return mock


@pytest.fixture
def transaction(accounts, networks):
    with networks.ethereum.local.use_provider("test"):
        sender = accounts.test_accounts[0]
        receiver = accounts.test_accounts[1]
        receipt = sender.transfer(receiver, "1 gwei")
        return receipt.transaction


@pytest.fixture
def txn_hash():
    return "0x55d07ce5e3f4f5742f3318cf328d700e43ee8cdb46000f2ac731a9379fca8ea7"


@pytest.fixture(
    params=(
        FEATURE_NOT_AVAILABLE_BECAUSE_OF_TIER_RESPONSE,
        FEATURE_NOT_AVAILABLE_BECAUSE_OF_NETWORK_RESPONSE,
    )
)
def feature_not_available_http_error(mocker, request):
    response = mocker.MagicMock(spec=Response)
    response.fixture_param = request.param  # For assertions
    response.json.return_value = {"error": {"message": request.param}}
    error = HTTPError(response=response)
    return error


@pytest.fixture
def alchemy_provider(networks) -> "Alchemy":
    return networks.ethereum.sepolia.get_provider("alchemy")



NETWORKS = {
    "abstract": [
        "testnet",
    ],
    "apechain": [
        "mainnet",
        "curtis",
    ],
    "arbitrum": [
        "mainnet",
        "nova",
        "sepolia",
    ],
    "astar": [
        "mainnet",
    ],
    "avalanche": [
        "mainnet",
        "fuji",
    ],
    "base": [
        "mainnet",
        "sepolia",
    ],
    "berachain": [
        "bartio",
    ],
    "blast": [
        "mainnet",
        "sepolia",
    ],
    "bsc": [
        "mainnet",
        "testnet",
        "opbnb",
        "opbnb-testnet",
    ],
    "crossfi": [
        "mainnet",
        "testnet",
    ],
    "ethereum": [
        "holesky",
        "mainnet",
        "sepolia",
    ],
    "fantom": [
        "opera",
        "testnet",
    ],
    "flow-evm": [
        "mainnet",
        "testnet",
    ],
    "geist": [
        "mainnet",
        "polter",
    ],
    "gnosis": [
        "mainnet",
        "chiado",
    ],
    "lens": [
        "sepolia",
    ],
    "linea": [
        "mainnet",
        "sepolia",
    ],
    "lumia": [
        "prism",
        "testnet",
    ],
    "mantle": [
        "mainnet",
        "sepolia",
    ],
    "metis": [
        "mainnet",
    ],
    "optimism": [
        "mainnet",
        "sepolia",
    ],
    "polygon": [
        "mainnet",
        "amoy",
    ],
    "polygon-zkevm": [
        "mainnet",
        "cardona",
    ],
    "rootstock": [
        "mainnet",
        "testnet",
    ],
    "scroll": [
        "mainnet",
        "sepolia",
    ],
    "shape": [
        "mainnet",
        "sepolia",
    ],
    "soneium": [
        "minato",
    ],
    "unichain": [
        "sepolia",
    ],
    "xmtp": [
        "sepolia",
    ],
    "world-chain": [
        "mainnet",
        "sepolia",
    ],
    "zetachain": [
        "mainnet",
        "testnet",
    ],
    "zksync": [
        "mainnet",
        "sepolia",
    ],
    "zora": [
        "mainnet",
        "sepolia",
    ],
}



import pytest
import websocket  # type: ignore
from ape import accounts, networks
from ape.exceptions import APINotImplementedError, TransactionNotFoundError
from ape.utils import ZERO_ADDRESS

from ape_alchemy._utils import NETWORKS
from ape_alchemy.provider import Alchemy


@pytest.fixture(params=[(name, net) for name, values in NETWORKS.items() for net in values])
def provider(request):
    ecosystem_cls = networks.get_ecosystem(request.param[0])
    network_cls = ecosystem_cls.get_network(request.param[1])
    with network_cls.use_provider("alchemy") as provider:
        yield provider


def test_http(provider):
    assert isinstance(provider, Alchemy)
    assert provider.http_uri.startswith("https")

    # NOTE: Sometimes the balance is 0, for some chains.
    assert provider.get_balance(ZERO_ADDRESS) is not None
    assert provider.get_block(0)
    assert provider.get_block("latest")


def test_ws(provider):
    ws_uri = provider.ws_uri
    if ws_uri is None:
        # Test will fail. Network does not support ws clients.
        return

    assert provider.ws_uri.startswith("wss")

    try:
        ws = websocket.WebSocket()
        ws.connect(provider.ws_uri)
        ws.close()

    except Exception as err:
        pytest.fail(f"Websocket URI not accessible. Reason: {err}")


def test_polygon_zkevm():
    # We noticed strange behavior on this network and thus called for more tests.
    with networks.polygon_zkevm.cardona.use_provider("alchemy") as provider:
        with pytest.raises(APINotImplementedError):
            _ = provider.priority_fee

        receiver = accounts.test_accounts[0]
        tx = provider.network.ecosystem.create_transaction(receiver=receiver)
        with pytest.raises(APINotImplementedError):
            _ = provider.create_access_list(tx)


def test_make_request_handles_result():
    """
    There was a bug where eth_call because ape-alchemy wasn't
    handling the result from make_request properly.
    """
    tx = {
        "to": "0x5576815a38A3706f37bf815b261cCc7cCA77e975",
        "value": "0x0",
        "data": "0x70a082310000000000000000000000005576815a38a3706f37bf815b261ccc7cca77e975",
    }
    with networks.polygon_zkevm.cardona.use_provider("alchemy") as provider:
        result = provider.make_request("eth_call", [tx, "latest"])
        assert not isinstance(result, dict)


def test_get_receipt():
    with networks.ethereum.sepolia.use_provider("alchemy") as alchemy:
        txn_hash = "0x68605140856c13038d325048c411aed98cc1eecc189f628a38edb597f6b9679e"
        existing_tx = alchemy.get_receipt(txn_hash)
        assert existing_tx.txn_hash == txn_hash

        txn_hash = "0x66600044856c13038d325048c411aed98cc1eecc189f628a38eeeeeeeeeeeeee"
        with pytest.raises(TransactionNotFoundError):
            _ = alchemy.get_receipt(txn_hash)



import os
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional

from ape.api import ReceiptAPI, TraceAPI, TransactionAPI, UpstreamProvider
from ape.exceptions import (
    APINotImplementedError,
    ContractLogicError,
    TransactionNotFoundError,
    VirtualMachineError,
)
from ape.logging import logger
from ape.utils import request_with_retry
from ape_ethereum.provider import Web3Provider
from eth_pydantic_types import HexBytes
from eth_typing import HexStr
from requests import HTTPError
from web3 import HTTPProvider, Web3
from web3.exceptions import ContractLogicError as Web3ContractLogicError
from web3.exceptions import ExtraDataLengthError, TransactionNotFound
from web3.gas_strategies.rpc import rpc_gas_price_strategy

try:
    from web3.middleware import ExtraDataToPOAMiddleware  # type: ignore
except ImportError:
    from web3.middleware import geth_poa_middleware as ExtraDataToPOAMiddleware  # type: ignore

from web3.middleware.validation import MAX_EXTRADATA_LENGTH
from web3.types import RPCEndpoint

from .exceptions import AlchemyFeatureNotAvailable, AlchemyProviderError, MissingProjectKeyError
from .trace import AlchemyTransactionTrace

if TYPE_CHECKING:
    from ape.types import BlockID
    from ape_ethereum.transactions import AccessList


# The user must either set one of these or an ENV VAR of the pattern:
#  WEB3_<ECOSYSTEM>_<NETWORK>_PROJECT_ID or  WEB3_<ECOSYSTEM>_<NETWORK>_API_KEY
DEFAULT_ENVIRONMENT_VARIABLE_NAMES = ("WEB3_ALCHEMY_PROJECT_ID", "WEB3_ALCHEMY_API_KEY")

# Alchemy will try to publish private transactions for 25 blocks.
PRIVATE_TX_BLOCK_WAIT = 25

# NOTE: "*" means "all networks".
NETWORKS_SUPPORTING_WEBSOCKETS = {
    "arbitrum": "*",
    "avalanche": "*",
    "base": "*",
    "bsc": ("mainnet", "testnet"),
    "berachain": "*",
    "blast": "*",
    "ethereum": "*",
    "fantom": "*",
    "geist": ("polter",),
    "gnosis": "*",
    "lens": "*",
    "linea": "*",
    "optimism": "*",
    "polygon": "*",
    "scroll": "*",
    "shape": "*",
    "soneium": "*",
    "unichain": "*",
    "world-chain": "*",
    "zksync": "*",
}


class Alchemy(Web3Provider, UpstreamProvider):
    """
    A web3 provider using an HTTP connection to Alchemy.

    Docs: https://docs.alchemy.com/alchemy/
    """

    network_uris: dict[tuple, str] = {}

    @property
    def uri(self):
        """
        Alchemy RPC URI, including the project ID.
        """
        ecosystem_name = self.network.ecosystem.name
        network_name = self.network.name
        if (ecosystem_name, network_name) in self.network_uris:
            return self.network_uris[(ecosystem_name, network_name)]

        key = None

        ecosystem_nm_part = ecosystem_name.upper().replace("-", "_")
        network_nm_part = network_name.upper().replace("-", "_")
        expected_env_var_prefix = f"WEB3_{ecosystem_nm_part}_{network_nm_part}_ALCHEMY"
        options = (
            *DEFAULT_ENVIRONMENT_VARIABLE_NAMES,
            f"{expected_env_var_prefix}_PROJECT_ID",
            f"{expected_env_var_prefix}_API_KEY",
        )

        for env_var_name in options:
            env_var = os.environ.get(env_var_name)
            if env_var:
                key = env_var
                break

        if not key:
            raise MissingProjectKeyError(options)

        network_name = self.network.name

        # NOTE: Fantom's mainnet is named "opera", but the Alchemy URI expects "mainnet".
        if self.network.ecosystem.name == "fantom" and self.network.name == "opera":
            network_name = "mainnet"

        default_format = "{0}-{1}.g.alchemy.com/v2/{2}"
        network_formats_by_ecosystem = {
            "arbitrum": "arb-{0}.g.alchemy.com/v2/{1}",
            "avalanche": "avax-{0}.g.alchemy.com/v2/{1}",
            "bsc": "bnb-{0}.g.alchemy.com/v2/{1}",
            "ethereum": "eth-{0}.g.alchemy.com/v2/{1}",
            "flow-evm": "flow-{0}.g.alchemy.com/v2/{1}",
            "fraxtal": "frax-{0}.g.alchemy.com/v2/{1}",
            "optimism": "opt-{0}.g.alchemy.com/v2/{1}",
            "polygon-zkevm": "polygonzkevm-{0}.g.alchemy.com/v2/{1}",
            "world-chain": "worldchain-{0}.g.alchemy.com/v2/{1}",
        }

        if ecosystem_name in network_formats_by_ecosystem:
            # Special cases.
            if network_name == "nova":
                uri = "arbnova-mainnet.g.alchemy.com/v2/{}".format(key)
            elif network_name.startswith("opbnb"):
                sub_network = "mainnet" if network_name == "opbnb" else "testnet"
                uri = "opbnb-{0}.g.alchemy.com/v2/{1}".format(sub_network, key)
            else:
                network_format = network_formats_by_ecosystem[ecosystem_name]
                uri = network_format.format(network_name, key)
        elif ecosystem_name == "xmtp" and network_name == "sepolia":
            uri = "xmtp-testnet.g.alchemy.com/v2/{0}".format(key)
        else:
            uri = default_format.format(ecosystem_name, network_name, key)

        uri = f"https://{uri}"
        self.network_uris[(ecosystem_name, network_name)] = uri
        return uri

    @property
    def http_uri(self) -> str:
        # NOTE: Overriding `Web3Provider.http_uri` implementation
        return self.uri

    @property
    def ws_uri(self) -> Optional[str]:
        ecosystem_name = self.network.ecosystem.name
        network_name = self.network.name
        supported_networks = NETWORKS_SUPPORTING_WEBSOCKETS.get(ecosystem_name, [])
        if supported_networks != "*" and network_name not in supported_networks:
            return None

        # NOTE: Overriding `Web3Provider.ws_uri` implementation
        return "ws" + self.uri[4:]  # Remove `http` in default URI w/ `ws`

    @property
    def priority_fee(self) -> int:
        if self.network.ecosystem.name == "polygon-zkevm":
            # The error is only 400 with no info otherwise.
            raise APINotImplementedError()

        return super().priority_fee

    @property
    def connection_str(self) -> str:
        return self.uri

    def connect(self):
        self._web3 = Web3(HTTPProvider(self.uri))
        is_poa = None
        try:
            # Any chain that *began* as PoA needs the middleware for pre-merge blocks
            base = 8453
            optimism = 10
            polygon = 137
            polygon_amoy = 80002

            if self._web3.eth.chain_id in (base, optimism, polygon, polygon_amoy):
                self._web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
                is_poa = True

            self._web3.eth.set_gas_price_strategy(rpc_gas_price_strategy)
        except Exception:
            is_poa = None

        if is_poa is None:
            # Check if is PoA but just wasn't as such yet.
            # NOTE: We have to check both earliest and latest
            #   because if the chain was _ever_ PoA, we need
            #   this middleware.
            for option in ("earliest", "latest"):
                try:
                    block = self.web3.eth.get_block(option)  # type: ignore[arg-type]
                except ExtraDataLengthError:
                    is_poa = True
                    break
                else:
                    is_poa = (
                        "proofOfAuthorityData" in block
                        or len(block.get("extraData", "")) > MAX_EXTRADATA_LENGTH
                    )
                    if is_poa:
                        break

            if is_poa and ExtraDataToPOAMiddleware not in self.web3.middleware_onion:
                self.web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

    def disconnect(self):
        self._web3 = None

    def _get_prestate_trace(self, transaction_hash: str) -> dict:
        return self.make_request(
            "debug_traceTransaction",
            [
                transaction_hash,
                {
                    "tracer": "prestateTracer",
                    "timeout": self.config.trace_timeout,
                },
            ],
        )

    def get_transaction_trace(self, transaction_hash: str, **kwargs) -> TraceAPI:
        return AlchemyTransactionTrace(transaction_hash=transaction_hash, **kwargs)

    def get_virtual_machine_error(self, exception: Exception, **kwargs) -> VirtualMachineError:
        txn = kwargs.get("txn")
        if not hasattr(exception, "args") or not len(exception.args):
            return VirtualMachineError(base_err=exception, txn=txn)

        args = exception.args
        message = args[0]
        if (
            not isinstance(exception, Web3ContractLogicError)
            and isinstance(message, dict)
            and "message" in message
        ):
            # Is some other VM error, like gas related
            return VirtualMachineError(message["message"], txn=txn)

        elif not isinstance(message, str):
            return VirtualMachineError(base_err=exception, txn=txn)

        # If get here, we have detected a contract logic related revert.
        message_prefix = "execution reverted"
        if message.startswith(message_prefix):
            message = message.replace(message_prefix, "")

            if ":" in message:
                # Was given a revert message
                message = message.split(":")[-1].strip()
                return ContractLogicError(revert_message=message, txn=txn)
            else:
                # No revert message
                return ContractLogicError(txn=txn)

        return VirtualMachineError(message=message, txn=txn)

    def create_access_list(
        self, transaction: TransactionAPI, block_id: Optional["BlockID"] = None
    ) -> list["AccessList"]:
        if self.network.ecosystem.name == "polygon-zkevm":
            # The error is only 400 with no info otherwise.
            raise APINotImplementedError()

        return super().create_access_list(transaction, block_id=block_id)

    def make_request(self, rpc: str, parameters: Optional[Iterable] = None) -> Any:
        rate_limit = self.config.rate_limit
        parameters = parameters or []
        try:
            result = request_with_retry(
                lambda: self.web3.provider.make_request(RPCEndpoint(rpc), parameters),
                min_retry_delay=rate_limit.min_retry_delay,
                retry_backoff_factor=rate_limit.retry_backoff_factor,
                max_retry_delay=rate_limit.max_retry_delay,
                max_retries=rate_limit.max_retries,
                retry_jitter=rate_limit.retry_jitter,
            )
        except HTTPError as err:
            response_data = err.response.json() if err.response else {}
            if "error" not in response_data:
                raise AlchemyProviderError(str(err)) from err

            error_data = response_data["error"]
            message = (
                error_data.get("message", str(error_data))
                if isinstance(error_data, dict)
                else error_data
            )
            cls = (
                AlchemyFeatureNotAvailable
                if "is not available" in message
                else AlchemyProviderError
            )
            raise cls(message) from err

        if isinstance(result, dict) and (res := result.get("result")):
            return res

        return result

    def send_private_transaction(self, txn: TransactionAPI, **kwargs) -> ReceiptAPI:
        """
        See `Alchemy's guide <https://www.alchemy.com/overviews/ethereum-private-transactions>`__
        for more information on sending private transaction using Alchemy.
        For more information on the API itself, see its
        `REST reference <https://docs.alchemy.com/reference/eth-sendprivatetransaction>`__.

        Args:
            txn: (:class:`~ape.api.transactionsTransactionAPI`): The transaction.
            **kwargs: Kwargs here are used for private-transaction "preferences".

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """
        max_block_number = kwargs.pop("max_block_number", None)

        params = {
            "tx": HexBytes(txn.serialize_transaction()).hex(),
            "maxBlockNumber": max_block_number,
        }
        if kwargs and "fast" not in kwargs:
            # If sending preferences, `fast` must be present.
            kwargs["fast"] = False
            params["preferences"] = kwargs

        try:
            txn_hash = self.make_request("eth_sendPrivateTransaction", [params])
        except (ValueError, Web3ContractLogicError) as err:
            vm_err = self.get_virtual_machine_error(err, txn=txn)
            raise vm_err from err

        # Since Alchemy will attempt to publish for 25 blocks,
        # we add 25 * block_time to the timeout.
        timeout = (
            PRIVATE_TX_BLOCK_WAIT * self.network.block_time
            + self.network.transaction_acceptance_timeout
        )

        receipt = self.get_receipt(
            txn_hash,
            required_confirmations=(
                txn.required_confirmations
                if txn.required_confirmations is not None
                else self.network.required_confirmations
            ),
            timeout=timeout,
        )
        logger.info(
            f"Confirmed {receipt.txn_hash} (private) (total fees paid = {receipt.total_fees_paid})"
        )
        self.chain_manager.history.append(receipt)
        return receipt

    def get_receipt(
        self,
        txn_hash: str,
        required_confirmations: int = 0,
        timeout: Optional[int] = None,
        **kwargs,
    ) -> ReceiptAPI:
        if not required_confirmations and not timeout:
            # Allows `get_receipt` to work better when not sending.
            try:
                data = self.web3.eth.get_transaction_receipt(HexStr(txn_hash))
            except TransactionNotFound as err:
                raise TransactionNotFoundError(txn_hash) from err

            txn = dict(self.web3.eth.get_transaction(HexStr(txn_hash)))
            return self.network.ecosystem.decode_receipt(
                {
                    "provider": self,
                    "required_confirmations": required_confirmations,
                    **txn,
                    **data,
                }
            )
        # Sending txns will get here because they always pass in required confs.
        return super().get_receipt(
            txn_hash, required_confirmations=required_confirmations, timeout=timeout, **kwargs
        )



from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from .config import AlchemyConfig

    return AlchemyConfig


@plugins.register(plugins.ProviderPlugin)
def providers():
    from ._utils import NETWORKS
    from .provider import Alchemy

    for ecosystem_name in NETWORKS:
        for network_name in NETWORKS[ecosystem_name]:
            yield ecosystem_name, network_name, Alchemy


def __getattr__(name: str):
    if name == "NETWORKS":
        from ._utils import NETWORKS

        return NETWORKS

    elif name == "Alchemy":
        from .provider import Alchemy

        return Alchemy

    elif name == "AlcheymyConfig":
        from .config import AlchemyConfig

        return AlchemyConfig

    raise AttributeError(name)


__all__ = [
    "NETWORKS",
    "Alchemy",
    "AlchemyConfig",
]



import re

import pytest
from ape.exceptions import ContractLogicError
from ape.types import LogFilter
from hexbytes import HexBytes
from requests import HTTPError
from web3.exceptions import ContractLogicError as Web3ContractLogicError

from ape_alchemy.provider import MissingProjectKeyError

TXN_HASH = "0x3cef4aaa52b97b6b61aa32b3afcecb0d14f7862ca80fdc76504c37a9374645c4"


@pytest.fixture
def parity_trace():
    return {
        "action": {
            "from": "0x5cab1e5286529370880776461c53a0e47d74fb63",
            "callType": "call",
            "gas": "0x17e6f0",
            "input": "0x96d373e5",
            "to": "0xc17f2c69ae2e66fd87367e3260412eeff637f70e",
            "value": "0x0",
        },
        "blockHash": "0xa7e0792b07687130af6042d9e295e7a96d83a34f40fe01074348cac5c5dd0699",
        "blockNumber": 15104985,
        "result": {"gasUsed": "0x1562f0", "output": "0x"},
        "subtraces": 1,
        "traceAddress": [],
        "transactionHash": TXN_HASH,
        "transactionPosition": 259,
        "type": "call",
    }


@pytest.fixture
def log_filter():
    return LogFilter(
        address=["0xF7F78379391C5dF2Db5B66616d18fF92edB82022"],
        fromBlock="0x3",
        toBlock="0x3",
        topics=[
            "0x1a7c56fae0af54ebae73bc4699b9de9835e7bb86b050dff7e80695b633f17abd",
            [
                "0x0000000000000000000000000000000000000000000000000000000000000000",
                "0x0000000000000000000000000000000000000000000000000000000000000001",
            ],
        ],
    )


@pytest.fixture
def block():
    return {
        "transactions": [],
        "hash": HexBytes("0xae1960ba0513948a507b652def457305d490d24bc0dd131d8d02be56564a3ee2"),
        "number": 0,
        "parentHash": HexBytes(
            "0x0000000000000000000000000000000000000000000000000000000000000000"
        ),
        "size": 517,
        "timestamp": 1660338772,
        "gasLimit": 30029122,
        "gasUsed": 0,
        "baseFeePerGas": 1000000000,
        "difficulty": 131072,
        "totalDifficulty": 131072,
    }


@pytest.fixture
def receipt():
    return {
        "blockNumber": 15329094,
        "data": b"0xa9059cbb00000000000000000000000016b308eb4591d9b4e34034ca2ff992d224b9927200000000000000000000000000000000000000000000000000000000030a32c0",  # noqa: E501
        "gasLimit": 79396,
        "gasPrice": 14200000000,
        "gasUsed": 65625,
        "logs": [
            {
                "blockHash": HexBytes(
                    "0x141a61b8c738c0f1508728116049a0d4a6ff41ee1180d956148880f32ae99215"
                ),
                "address": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
                "logIndex": 213,
                "data": HexBytes(
                    "0x00000000000000000000000000000000000000000000000000000000030a32c0"
                ),
                "removed": False,
                "topics": [
                    HexBytes("0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"),
                    HexBytes("0x000000000000000000000000958f973513f723f2cb9b47abe5e903695ab93e36"),
                    HexBytes("0x00000000000000000000000016b308eb4591d9b4e34034ca2ff992d224b99272"),
                ],
                "blockNumber": 15329094,
                "transactionIndex": 132,
                "transactionHash": HexBytes(
                    "0x9e4be62c1a16caacaccd9d8c7706b75dc17a957ec6c5dea418a137a5c3a197c5"
                ),
            }
        ],
        "nonce": 16,
        "receiver": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
        "sender": "0x958f973513F723f2cB9b47AbE5e903695aB93e36",
        "status": 1,
        "hash": TXN_HASH,
        "value": 0,
    }


def test_when_no_api_key_raises_error(missing_token, alchemy_provider):
    with pytest.raises(
        MissingProjectKeyError,
        match=re.escape(
            "Must set one of "
            "$WEB3_ALCHEMY_PROJECT_ID, "
            "$WEB3_ALCHEMY_API_KEY, "
            "$WEB3_ETHEREUM_SEPOLIA_ALCHEMY_PROJECT_ID, "
            "$WEB3_ETHEREUM_SEPOLIA_ALCHEMY_API_KEY."
        ),
    ):
        alchemy_provider.connect()


def test_send_transaction_reverts(token, alchemy_provider, mock_web3, transaction):
    expected_revert_message = "EXPECTED REVERT MESSAGE"
    mock_web3.eth.send_raw_transaction.side_effect = Web3ContractLogicError(
        f"execution reverted : {expected_revert_message}"
    )
    alchemy_provider._web3 = mock_web3

    with pytest.raises(ContractLogicError, match=expected_revert_message):
        alchemy_provider.send_transaction(transaction)


def test_send_transaction_reverts_no_message(token, alchemy_provider, mock_web3, transaction):
    mock_web3.eth.send_raw_transaction.side_effect = Web3ContractLogicError("execution reverted")
    alchemy_provider._web3 = mock_web3

    with pytest.raises(ContractLogicError, match="Transaction failed."):
        alchemy_provider.send_transaction(transaction)


def test_estimate_gas_would_revert(token, alchemy_provider, mock_web3, transaction):
    expected_revert_message = "EXPECTED REVERT MESSAGE"
    mock_web3.eth.estimate_gas.side_effect = Web3ContractLogicError(
        f"execution reverted : {expected_revert_message}"
    )
    alchemy_provider._web3 = mock_web3

    with pytest.raises(ContractLogicError, match=expected_revert_message):
        alchemy_provider.estimate_gas_cost(transaction)


def test_estimate_gas_would_revert_no_message(token, alchemy_provider, mock_web3, transaction):
    mock_web3.eth.estimate_gas.side_effect = Web3ContractLogicError("execution reverted")
    alchemy_provider._web3 = mock_web3

    with pytest.raises(ContractLogicError, match="Transaction failed."):
        alchemy_provider.estimate_gas_cost(transaction)


def test_get_contract_logs(networks, alchemy_provider, mock_web3, block, log_filter):
    mock_web3.eth.get_block.return_value = block
    alchemy_provider._web3 = mock_web3
    networks.active_provider = alchemy_provider
    actual = [x for x in alchemy_provider.get_contract_logs(log_filter)]

    # Fails when improper response handling of logs (is part of bug fix)
    assert actual == []


def test_get_transaction_trace(networks, alchemy_provider, mock_web3, parity_trace, receipt):
    mock_web3.provider.make_request.return_value = [parity_trace]
    mock_web3.eth.wait_for_transaction_receipt.return_value = receipt
    alchemy_provider._web3 = mock_web3
    networks.active_provider = alchemy_provider
    trace = alchemy_provider.get_transaction_trace(TXN_HASH)
    actual = repr(trace.get_calltree())
    expected = r"CALL: 0xC17f2C69aE2E66FD87367E3260412EEfF637F70E\.<0x96d373e5\> \[1401584 gas\]"
    assert re.match(expected, actual)


def test_make_request_rate_limiting(mocker, alchemy_provider, mock_web3):
    alchemy_provider._web3 = mock_web3

    class RateLimitTester:
        tries = 2
        _try = 0
        tries_made = 0

        def rate_limit_hook(self, rpc, params):
            self.tries_made += 1
            if self._try == self.tries:
                self._try = 0
                return {"success": True}
            else:
                self._try += 1
                response = mocker.MagicMock()
                response.status_code = 429
                raise HTTPError(response=response)

    rate_limit_tester = RateLimitTester()
    mock_web3.provider.make_request.side_effect = rate_limit_tester.rate_limit_hook
    result = alchemy_provider.make_request("ape_testRateLimiting", parameters=[])
    assert rate_limit_tester.tries_made == rate_limit_tester.tries + 1
    assert result == {"success": True}



import pytest
from ape import chain, networks
from ethpm_types import ContractType


@pytest.fixture(autouse=True)
def ethereum_mainnet_alchemy():
    with networks.ethereum.mainnet.use_provider("alchemy"):
        yield


def test_revert_message():
    txn_hash = "0x36144f609e0fc7afd3cc570d6a54582091642a44c5223a5ad59aa20008dd9577"
    receipt = chain.history[txn_hash]
    actual = receipt.trace.revert_message
    expected = "UniswapV2Router: INSUFFICIENT_OUTPUT_AMOUNT"
    assert actual == expected


def test_return_value():
    txn_hash = "0xe0897d735b67893648b20085ecef16232733425329df844292d5b2774cca436b"
    receipt = chain.history[txn_hash]

    # Ensure the ABI is cached so we can decode the return value.
    abi = [
        {
            "type": "function",
            "name": "submit",
            "stateMutability": "payable",
            "inputs": [{"name": "_referral", "type": "address"}],
            "outputs": [{"name": "", "type": "uint256"}],
        }
    ]
    chain.contracts[receipt.receiver] = ContractType(abi=abi)

    actual = receipt.return_value
    expected = 1244617160572980465
    assert actual == expected



from functools import cached_property
from typing import Any, Optional

from ape_ethereum.trace import TraceApproach, TransactionTrace
from hexbytes import HexBytes


class AlchemyTransactionTrace(TransactionTrace):
    call_trace_approach: TraceApproach = TraceApproach.PARITY

    @cached_property
    def return_value(self) -> Any:
        node = self._top_level_call
        if output := node.get("output"):
            output_bytes = HexBytes(output)
            if abi := self.root_method_abi:
                return self._ecosystem.decode_returndata(abi, output_bytes)

            # ABI is not known.
            return output_bytes

        return None

    @cached_property
    def revert_message(self) -> Optional[str]:
        node = self._top_level_call
        return node.get("revertReason")

    @cached_property
    def _top_level_call(self) -> dict:
        return self.provider.make_request(
            "debug_traceTransaction",
            [
                self.transaction_hash,
                {
                    "tracer": "callTracer",
                    "timeout": self.provider.config.trace_timeout,
                    "tracerConfig": {
                        "onlyTopLevelCall": True,
                    },
                },
            ],
        )
